r"""Doctests in TestCase classes

The original ``doctest`` unittest integration was based on
``unittest`` test suites, which have fallen out of favor. This module
provides a way to define doctests inside of unittest ``TestCase``
classes. It also provides better integration with unittest test
fixtures, because doctests use setup provided by the containing test
case class.  It also provides access to unittest assertion
methods.

You can define doctests in 4 ways:

- references to named files

- strings

- decorated functions with docstrings

- reference to named files decorating test-specific setup functions

.. some setup

   >>> __name__ = 'tests'

Here are some examples::

    from zope.testing import doctestcase
    import doctest
    import unittest

    g = 'global'
    class MyTest(unittest.TestCase):

        def setUp(self):
            self.a = 1
            self.globs = dict(c=9)

        test1 = doctestcase.file('test1.txt', optionflags=doctest.ELLIPSIS)

        test2 = doctestcase.docteststring('''
          >>> self.a, g, c
          (1, 'global', 9)
        ''')

        @doctestcase.doctestmethod(optionflags=doctest.ELLIPSIS)
        def test3(self):
            '''
            >>> self.a, self.x, g, c
            (1, 3, 'global', 9)
            '''
            self.x = 3

        @doctestcase.doctestfile('test4.txt')
        def test4(self):
            self.x = 5

In this example, 3 constructors were used:

doctestfile (alias: file)
  doctestfile makes a file-based test case.

  This can be used as a decorator, in which case, the decorated
  function is called before the test is run, to provide test-specific
  setup.

docteststring (alias string)
  docteststring constructs a doctest from a string.

doctestmethod (alias method)
  doctestmethod constructs a doctest from a method.

  The method's docstring provides the test. The method's body provides
  optional test-specific setup.

Note that short aliases are provided, which may be useful in certain
import styles.

Tests have access to the following data:

- Tests created with the ``docteststring`` and ``doctestmethod``
  constructors have access to the module globals of the defining
  module.

- In tests created with the ``docteststring`` and ``doctestmethod``
  constructors, the test case instance is available as the ``self``
  variable.

- In tests created with the ``doctestfile`` constructor, the test case
  instance is available as the ``test`` variable.

- If a test case defines a globs attribute, it must be a dictionary
  and its contents are added to the test globals.

The constructors accept standard doctest ``optionflags`` and
``checker`` arguments.

Note that the doctest IGNORE_EXCEPTION_DETAIL option flag is
added to optionflags.
"""
import doctest
import inspect
import os
import re
import sys
import types

__all__ = ['doctestmethod', 'docteststring', 'doctestfile']

_parser = doctest.DocTestParser()


def _testify(name):
    if not name.startswith('test'):
        name = 'test_' + name
    return name


def doctestmethod(test=None, optionflags=0, checker=None):
    """Define a doctest from a method within a unittest.TestCase.

    The method's doc string provides the test source. Its body is
    called before the test and may perform test-specific setup.

    You can pass doctest option flags and a custon checker.

    Variables defined in the enclosing module are available in the test.

    If a test case defines a globs attribute, it must be a dictionary
    and its contents are added to the test globals.

    The test object is available as the variable ``self`` in the test.
    """
    if test is None:
        return lambda test: _doctestmethod(test, optionflags, checker)

    return _doctestmethod(test, optionflags, checker)


method = doctestmethod


def _doctestmethod(test, optionflags, checker):
    doc = test.__doc__
    if not doc:
        raise ValueError(test, "has no docstring")
    setup = test
    name = test.__name__
    path = inspect.getsourcefile(test)
    lineno = inspect.getsourcelines(test)[1]

    fglobs = sys._getframe(3).f_globals

    def test_method(self):
        setup(self)
        _run_test(self, doc, fglobs.copy(), name, path,
                  optionflags, checker, lineno=lineno)

    test_method.__name__ = _testify(name)

    return test_method


def docteststring(test, optionflags=0, checker=None, name=None):
    """Define a doctest from a string within a unittest.TestCase.

    You can pass doctest option flags and a custon checker.

    Variables defined in the enclosing module are available in the test.

    If a test case defines a globs attribute, it must be a dictionary
    and its contents are added to the test globals.

    The test object is available as the variable ``self`` in the test.
    """
    fglobs = sys._getframe(2).f_globals

    def test_string(self):
        _run_test(self, test, fglobs.copy(), '<string>', '<string>',
                  optionflags, checker)
    if name:
        test_string.__name__ = _testify(name)

    return test_string


string = docteststring

_not_word = re.compile(r'\W')


def doctestfile(path, optionflags=0, checker=None):
    """Define a doctest from a test file within a unittest.TestCase.

    The file path may be relative or absolute. If its relative (the
    common case), it will be interpreted relative to the directory
    containing the referencing module.

    You can pass doctest option flags and a custon checker.

    If a test case defines a globs attribute, it must be a dictionary
    and its contents are added to the test globals.

    The test object is available as the variable ``test`` in the test.

    The resulting object can be used as a function decorator. The
    decorated method is called before the test and may perform
    test-specific setup. (The decorated method's doc string is ignored.)
    """
    base = os.path.dirname(os.path.abspath(
        sys._getframe(2).f_globals['__file__']
        ))
    path = os.path.join(base, path)
    with open(path) as f:
        test = f.read()
    name = os.path.basename(path)

    def test_file(self):
        if isinstance(self, types.FunctionType):
            setup = self

            def test_file_w_setup(self):
                setup(self)
                _run_test(self, test, {}, name, path, optionflags, checker,
                          'test')

            test_file_w_setup.__name__ = _testify(setup.__name__)
            test_file_w_setup.filepath = path
            test_file_w_setup.filename = os.path.basename(path)
            return test_file_w_setup

        _run_test(self, test, {}, name, path, optionflags, checker, 'test')

    test_file.__name__ = name_from_path(path)
    test_file.filepath = path
    test_file.filename = os.path.basename(path)

    return test_file


file = doctestfile


def doctestfiles(*paths, **kw):
    """Define doctests from test files in a decorated class.

    Multiple files can be specified. A member is added to the
    decorated class for each file.

    The file paths may be relative or absolute. If relative (the
    common case), they will be interpreted relative to the directory
    containing the referencing module.

    You can pass doctest option flags and a custon checker.

    If a test case defines a globs attribute, it must be a dictionary
    and its contents are added to the test globals.

    The test object is available as the variable ``test`` in the test.

    The resulting object must be used as a class decorator.
    """
    def doctestfiles_(class_):
        for path in paths:
            name = name_from_path(path)
            test = doctestfile(path, **kw)
            test.__name__ = name
            setattr(class_, name, test)

        return class_

    return doctestfiles_


files = doctestfiles


def name_from_path(path):
    return _testify(
        _not_word.sub('_', os.path.splitext(os.path.basename(path))[0])
        )


def _run_test(self, test, globs, name, path,
              optionflags, checker, testname='self', lineno=0):
    globs.update(getattr(self, 'globs', ()))
    globs[testname] = self
    optionflags |= doctest.IGNORE_EXCEPTION_DETAIL
    doctest.DocTestCase(
        _parser.get_doctest(test, globs, name, path, lineno),
        optionflags=optionflags,
        checker=checker,
        ).runTest()



import sys
import textwrap
import unittest

from zope.testing.renormalizing import (
    is_dotted_name,
    strip_dottedname_from_traceback,
)


class Exception2To3(unittest.TestCase):

    def test_is_dotted_name_ascii_no_dots(self):
        self.assertTrue(is_dotted_name('FooBarError'))

    def test_is_dotted_name_ascii_dots(self):
        self.assertTrue(is_dotted_name('foo.bar.FooBarError'))

    def test_is_dotted_name_unicode_no_dots(self):
        result = is_dotted_name(u'FooB\xe1rError')
        if sys.version_info[0] >= 3:
            self.assertTrue(result)
        else:
            self.assertFalse(result)

    def test_is_dotted_name_unicode_dots(self):
        result = is_dotted_name(u'foo.b\xe1r.FooB\xe1rError')
        if sys.version_info[0] >= 3:
            self.assertTrue(result)
        else:
            self.assertFalse(result)

    def test_is_dotted_name_ellipsis(self):
        self.assertFalse(is_dotted_name('...'))

    def test_is_dotted_name_not_identifier(self):
        self.assertFalse(is_dotted_name('foo=bar'))

    def test_strip_dottedname(self):
        string = textwrap.dedent("""\
            Traceback (most recent call last):
            foo.bar.FooBarError: requires at least one argument.""")
        expected = textwrap.dedent("""\
            Traceback (most recent call last):
            FooBarError: requires at least one argument.""")
        self.assertEqual(expected, strip_dottedname_from_traceback(string))

    def test_strip_dottedname_without_exception_arguments(self):
        string = textwrap.dedent("""\
            Traceback (most recent call last):
            foo.bar.FooBarError""")
        expected = textwrap.dedent("""\
            Traceback (most recent call last):
            FooBarError""")
        self.assertEqual(expected, strip_dottedname_from_traceback(string))

    def test_no_dots_in_name(self):
        string = textwrap.dedent("""\
            Traceback (most recent call last):
            FooBarError: requires at least one argument.""")
        expected = textwrap.dedent("""\
            Traceback (most recent call last):
            FooBarError: requires at least one argument.""")
        self.assertEqual(expected, strip_dottedname_from_traceback(string))

    def test_no_colon_in_first_word(self):
        string = textwrap.dedent("""\
            Traceback (most recent call last):
            foo.bar.FooBarError requires at least one argument.""")
        expected = textwrap.dedent("""\
            Traceback (most recent call last):
            foo.bar.FooBarError requires at least one argument.""")
        self.assertEqual(expected, strip_dottedname_from_traceback(string))

    def test_input_empty(self):
        string = ''
        expected = ''
        self.assertEqual(expected, strip_dottedname_from_traceback(string))

    def test_input_spaces(self):
        string = '   '
        expected = '   '
        self.assertEqual(expected, strip_dottedname_from_traceback(string))

    def test_input_ellipsis(self):
        string = '...'
        expected = '...'
        self.assertEqual(expected, strip_dottedname_from_traceback(string))

    def test_input_last_line_not_dotted_name(self):
        string = 'foo=bar'
        expected = 'foo=bar'
        self.assertEqual(expected, strip_dottedname_from_traceback(string))

    def test_last_line_empty(self):
        string = textwrap.dedent("""\
            Traceback (most recent call last):

            """)
        expected = textwrap.dedent("""\
            Traceback (most recent call last):

            """)
        self.assertEqual(expected, strip_dottedname_from_traceback(string))



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""logging handler for tests that check logging output.
"""
import logging


class Handler(logging.Handler):
    """Handler for use with unittest.TestCase objects.

    The handler takes a TestCase instance as a constructor argument.
    It can be registered with one or more loggers and collects log
    records they generate.

    The assertLogsMessage() and failIfLogsMessage() methods can be
    used to check the logger output and causes the test to fail as
    appropriate.
    """

    def __init__(self, testcase, propagate=False):
        logging.Handler.__init__(self)
        self.records = []
        # loggers stores (logger, propagate) tuples
        self.loggers = []
        self.closed = False
        self.propagate = propagate
        self.testcase = testcase

    def close(self):
        """Remove handler from any loggers it was added to."""
        if self.closed:
            return
        for logger, propagate in self.loggers:
            logger.removeHandler(self)
            logger.propagate = propagate
        self.closed = True

    def add(self, name):
        """Add handler to logger named name."""
        logger = logging.getLogger(name)
        old_prop = logger.propagate
        logger.addHandler(self)
        if self.propagate:
            logger.propagate = 1
        else:
            logger.propagate = 0
        self.loggers.append((logger, old_prop))

    def emit(self, record):
        self.records.append(record)

    def assertLogsMessage(self, msg, level=None):
        for r in self.records:
            if r.getMessage() == msg:
                if level is not None and r.levelno == level:
                    return
        msg = "No log message contained %r" % msg
        if level is not None:
            msg += " at level %d" % level
        self.testcase.fail(msg)

    def failIfLogsMessage(self, msg):
        for r in self.records:
            if r.getMessage() == msg:
                self.testcase.fail("Found log message %r" % msg)



##############################################################################
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests for the testing framework.
"""
import doctest
import sys
import re
import unittest
from zope.testing import renormalizing
from zope.testing.test_renormalizing import Exception2To3


def print_(*args):
    sys.stdout.write(' '.join(map(str, args))+'\n')


def setUp(test):
    test.globs['print_'] = print_


def test_suite():
    suite = unittest.TestSuite((
        doctest.DocFileSuite(
            'module.txt',
            # Python 3.3 changed exception messaging:
            #   https://bugs.launchpad.net/zope.testing/+bug/1055720
            # and then Python 3.6 introduced ImportError subclasses
            checker=renormalizing.RENormalizing([
                (re.compile('ModuleNotFoundError:'), 'ImportError:'),
                (re.compile(
                    "No module named '?zope.testing.unlikelymodulename'?"),
                 'No module named unlikelymodulename'),
                (re.compile("No module named '?fake'?"),
                 'No module named fake')])),
        doctest.DocFileSuite('loggingsupport.txt', setUp=setUp),
        doctest.DocFileSuite('renormalizing.txt', setUp=setUp),
        doctest.DocFileSuite('setupstack.txt', setUp=setUp),
        doctest.DocFileSuite(
            'wait.txt', setUp=setUp,
            checker=renormalizing.RENormalizing([
                # For Python 3.4.
                (re.compile('zope.testing.wait.TimeOutWaitingFor: '),
                 'TimeOutWaitingFor: '),
                # For Python 3.5
                (re.compile('zope.testing.wait.Wait.TimeOutWaitingFor: '),
                 'TimeOutWaitingFor: '),
                ])
            ),
        ))

    if sys.version_info[:2] >= (2, 7):
        suite.addTests(doctest.DocFileSuite('doctestcase.txt'))
    if sys.version_info[0] < 3:
        suite.addTests(doctest.DocTestSuite('zope.testing.server'))
        suite.addTests(doctest.DocFileSuite('formparser.txt'))
    suite.addTest(
        unittest.makeSuite(Exception2To3))
    return suite



##############################################################################
#
# Copyright (c) 2005 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Stack-based test doctest setUp and tearDown

See setupstack.txt
"""

import os
import stat
import tempfile
import unittest


key = '__' + __name__


def globs(test):
    try:
        return test.globs
    except AttributeError:
        return test.__dict__


def register(test, function, *args, **kw):
    tglobs = globs(test)
    stack = tglobs.get(key)
    if stack is None:
        stack = tglobs[key] = []
    stack.append((function, args, kw))


def tearDown(test):
    tglobs = globs(test)
    stack = tglobs.get(key)
    while stack:
        f, p, k = stack.pop()
        f(*p, **k)


def setUpDirectory(test):
    tmp = tempfile.mkdtemp()
    register(test, rmtree, tmp)
    here = os.getcwd()
    register(test, os.chdir, here)
    os.chdir(tmp)


def rmtree(path):
    for path, dirs, files in os.walk(path, False):
        for fname in files:
            fname = os.path.join(path, fname)
            if not os.path.islink(fname):
                os.chmod(fname, stat.S_IWUSR)
            os.remove(fname)
        for dname in dirs:
            dname = os.path.join(path, dname)
            os.rmdir(dname)
    os.rmdir(path)


def context_manager(test, manager):
    result = manager.__enter__()
    register(test, manager.__exit__, None, None, None)
    return result


def mock(test, *args, **kw):
    import mock as mock_module
    return context_manager(test, mock_module.patch(*args, **kw))


class TestCase(unittest.TestCase):

    tearDown = tearDown
    register = register
    setUpDirectory = setUpDirectory
    context_manager = context_manager
    mock = mock



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Provide a standard cleanup registry

Unit tests that change global data should include the CleanUp base
class, which provides simpler setUp and tearDown methods that call
global-data cleanup routines::

  class Test(CleanUp, unittest.TestCase):

      ....

If custom setUp or tearDown are needed, then the base routines should
be called, as in::

  def tearDown(self):
      super(Test, self).tearDown()
      ....

Cleanup routines for global data should be registered by passing them to
addCleanup::


  addCleanUp(pigRegistry._clear)

"""

_cleanups = []


def addCleanUp(func, args=(), kw={}):
    """Register a cleanup routines

    Pass a function to be called to cleanup global data.
    Optional argument tuple and keyword arguments may be passed.
    """
    _cleanups.append((func, args, kw))


class CleanUp(object):
    """Mix-in class providing clean-up setUp and tearDown routines."""

    def cleanUp(self):
        """Clean up global data."""
        cleanUp()

    setUp = tearDown = cleanUp


def cleanUp():
    """Clean up global data."""
    for func, args, kw in _cleanups:
        func(*args, **kw)


setUp = tearDown = cleanUp



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest
import re
import sys


IGNORE_EXCEPTION_MODULE_IN_PYTHON2 = doctest.register_optionflag(
    'IGNORE_EXCEPTION_MODULE_IN_PYTHON2')
IGNORE_EXCEPTION_MODULE_IN_PYTHON2_HINT = """\
===============================================================
HINT:
  The optionflag IGNORE_EXCEPTION_MODULE_IN_PYTHON2 is set.
  You seem to test traceback output.
  If you are indeed, make sure to use the full dotted name of
  the exception class like Python 3 displays,
  even though you are running the tests in Python 2.
  The exception message needs to be last line (and thus not
  split over multiple lines).
==============================================================="""


class OutputChecker(doctest.OutputChecker):
    """Pattern-normalizing output checker
    """

    def __init__(self, patterns=None):
        if patterns is None:
            patterns = []
        self.transformers = list(map(self._cook, patterns))

    def __add__(self, other):
        if not isinstance(other, RENormalizing):
            return NotImplemented
        return RENormalizing(self.transformers + other.transformers)

    def _cook(self, pattern):
        if hasattr(pattern, '__call__'):
            return pattern
        regexp, replacement = pattern
        return lambda text: regexp.sub(replacement, text)

    def check_output(self, want, got, optionflags):
        if got == want:
            return True

        for transformer in self.transformers:
            want = transformer(want)
            got = transformer(got)

        if doctest.OutputChecker.check_output(self, want, got, optionflags):
            return True

        if sys.version_info[0] < 3:
            if optionflags & IGNORE_EXCEPTION_MODULE_IN_PYTHON2:
                want = strip_dottedname_from_traceback(want)
                if doctest.OutputChecker.check_output(
                        self, want, got, optionflags):
                    return True

        return False

    def output_difference(self, example, got, optionflags):

        want = example.want

        # If want is empty, use original outputter. This is useful
        # when setting up tests for the first time.  In that case, we
        # generally use the differencer to display output, which we evaluate
        # by hand.
        if not want.strip():
            return doctest.OutputChecker.output_difference(
                self, example, got, optionflags)

        # Dang, this isn't as easy to override as we might wish
        original = want

        for transformer in self.transformers:
            want = transformer(want)
            got = transformer(got)

        # temporarily hack example with normalized want:
        example.want = want

        if sys.version_info[0] < 3:
            if optionflags & IGNORE_EXCEPTION_MODULE_IN_PYTHON2:
                if maybe_a_traceback(got) is not None:
                    got += IGNORE_EXCEPTION_MODULE_IN_PYTHON2_HINT

        result = doctest.OutputChecker.output_difference(
            self, example, got, optionflags)
        example.want = original

        return result


RENormalizing = OutputChecker


def is_dotted_name(name):
    if sys.version_info[0] >= 3:
        return (
            name and
            all(element.isidentifier() for element in name.split('.')))
    else:
        # Python 2 lacked str.isidentifier, but also restricted identifiers
        # to ASCII so a regex match is straightforward.
        match = re.match(
            r'^(?:[A-Za-z_][A-Za-z0-9_]*\.)*[A-Za-z_][A-Za-z0-9_]*$', name)
        return match is not None


def maybe_a_traceback(string):
    # We wanted to confirm more strictly we're dealing with a traceback here.
    # However, doctest will preprocess exception output. It gets rid of the
    # the stack trace and the "Traceback (most recent call last)"-part. It
    # passes only the exception message to the checker.
    if not string:
        return None

    lines = string.splitlines()
    last = lines[-1]
    if not last:
        return None
    words = last.split(' ')
    first = words[0]
    if len(words) > 1 and not first.endswith(':'):
        return None
    # If IGNORE_EXCEPTION_MODULE_IN_PYTHON2 was applied to an entire file,
    # then this may run on strings that aren't the exception message part of
    # a traceback.  The doctest interface makes it impossible to detect this
    # reasonably, so do our best to restrict this to only lines that start
    # with something that looks like a Python dotted name.  It's best to
    # apply IGNORE_EXCEPTION_MODULE_IN_PYTHON2 only to examples that need
    # it.
    if not is_dotted_name(first[:-1]):
        return None

    return lines, last, words, first


def strip_dottedname_from_traceback(string):
    maybe = maybe_a_traceback(string)
    if maybe is None:
        return string

    lines, last, words, first = maybe
    name = first.split('.')[-1]
    words[0] = name
    last = ' '.join(words)
    lines[-1] = last
    return '\n'.join(lines)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Functional test server to interactively inspect the state of the system.

You can run it in a functional test by adding a line like this:

  startServer(http, url, "username", "password")

http is an instance of HTTPCaller, url is the url that will be opened
in the browser, the username and password are optional. When you're
done with inspecting the application press Ctrl+C to continue with the
functional test.
"""
from __future__ import print_function

import sys
import webbrowser
import urlparse
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
# XXX: I don't think this module works on Python 3!


def makeRequestHandler(http, user=None, password=None):
    class FunctionalTestRequestHandler(BaseHTTPRequestHandler):

        def do_GET(self):
            request = self.raw_requestline
            if user and password:
                # Authentication is built in, as there is no fluent
                # way of transferring session from functional test to
                # the real browser
                request += "Authorization: Basic %s:%s\r\n" % (user, password)

            # Write headers to the request
            for header in self.headers.headers:
                request += header
            request += '\r\n'

            if self.headers.get('Content-Length'):
                data = self.rfile.read(int(self.headers.get('Content-Length')))
                request += data
            else:
                # if no content-length was set - read until the last
                # char, then finish
                self.request.setblocking(0)
                while True:
                    try:
                        char = self.rfile.read()
                    except Exception:  # XXX: should probably be IOError?
                        break
                    request += char

            response = http(request)
            self.wfile.write(response)

        do_POST = do_GET

    return FunctionalTestRequestHandler


def addPortToURL(url, port):
    """Add a port number to the url.

        >>> addPortToURL('http://localhost/foo/bar/baz.html', 3000)
        'http://localhost:3000/foo/bar/baz.html'
        >>> addPortToURL('http://foo.bar.com/index.html?param=some-value', 555)
        'http://foo.bar.com:555/index.html?param=some-value'

        >>> addPortToURL('http://localhost:666/index.html', 555)
        'http://localhost:555/index.html'

    """
    (scheme, netloc, url, query, fragment) = urlparse.urlsplit(url)
    netloc = netloc.split(':')[0]
    netloc = "%s:%s" % (netloc, port)
    url = urlparse.urlunsplit((scheme, netloc, url, query, fragment))
    return url


def startServer(http, url, user=None, password=None, port=8000):
    try:
        server_address = ('', port)
        requestHandler = makeRequestHandler(http, user, password)
        url = addPortToURL(url, port)
        httpd = HTTPServer(server_address, requestHandler)
        # XXX we rely on browser being slower than our server
        webbrowser.open(url)
        print('Starting HTTP server...', file=sys.stderr)
        httpd.serve_forever()
    except KeyboardInterrupt:
        print('Stopped HTTP server.', file=sys.stderr)
